/////////////////////////////////
//    Compute RKAT/DKATfam/DKAT 
/////////////////////////////////

#include <Eigen/Dense>

#include <boost/math/distributions/beta.hpp>
#include <boost/math/distributions/binomial.hpp>
#include <boost/math/distributions/normal.hpp>
#include <boost/math/distributions/gamma.hpp>
#include <algorithm> 
#include <iostream>
#include <fstream>

using Eigen::MatrixXd;
using namespace Eigen;
using namespace boost::math;
using namespace std;

typedef Map<Matrix<double, Dynamic, Dynamic, RowMajor> > MapRowXd;
typedef Map<Matrix<int, Dynamic, Dynamic, RowMajor> > MapRowXi;
typedef Map<ArrayXXd> MapArXd;

#include "jasper.h"
double pearson_pval(const Ref<const MatrixXd>& Ymat, const Ref<const MatrixXd>& Gmat, int niter);
void get_moments(const Ref<const MatrixXd>& Aker, const Ref<const MatrixXd>& Wker, struct PARAMS *params);
void get_quantities(const Ref<const MatrixXd>& A, struct QUANTS* qA);
double mc_iter(const Ref<const MatrixXd>& Yker, const Ref<const MatrixXd>& Gmat, double Qobs, int n_iter);


int main(int argc, char **argv){

    char *end = NULL, pheno_filename[MAXFILELEN], covG_filename[MAXFILELEN], geno_filename[MAXFILELEN], grm_filename[MAXFILELEN], eigen_filename[MAXFILELEN], out_prefix[MAXFILELEN], eigout_filename[MAXFILELEN], pval_filename[MAXFILELEN];
    int arg, ofile = 0, cfile=0, gfile=0, efile=0, print_eigen=0, n_total=0, n_pheno=0, nmarker=0, niter=0;
    long niter_long = 0;
    double  time_JASPER;
    clock_t time1, time2;
    time1 = clock();


    printf("\n"
            "=========================================================\n"
            "|                          JASPER                       |\n"
            "|    Joint Association analysis in Structured samples   |\n"
            "|    based on approximating a PERmutation distribution  |\n"
            "|                                                       |\n"
            "|                                                       |\n"
            "|    (c) 2019-2023 Joelle Mbatchou, Mary Sara McPeek    |\n"
            "|                                                       |\n"
            "=========================================================\n"
            "\n"
          );

    for (int arg = 1; arg < argc; arg++) {
        if (strcmp(argv[arg], "--pheno") == 0){
            if (argv[arg + 1] == NULL || argv[arg + 1][0] == '-') {
                continue;
            }
            arg++;
            pheno_filename[MAXFILELEN-1]='\0';
            strncpy(pheno_filename, argv[arg], MAXFILELEN);
            if(pheno_filename[MAXFILELEN-1]!='\0'){
                printf("ERROR: Input file names cannot exceed %d characters. Please double check the name of the trait file!\n", MAXFILELEN-1);
                exit(1);
            }
        } else if (strcmp(argv[arg], "--geno") == 0){
            if (argv[arg + 1] == NULL || argv[arg + 1][0] == '-') {
                continue;
            }
            arg++;
            geno_filename[MAXFILELEN-1]='\0';
            strncpy(geno_filename, argv[arg], MAXFILELEN);
            if(geno_filename[MAXFILELEN-1]!='\0'){
                printf("ERROR: Input file names cannot exceed %d characters. Please double check the name of the genotype file!\n", MAXFILELEN-1);
                exit(1);
            }
        } else if (strcmp(argv[arg], "--covG") == 0){
            if (argv[arg + 1] == NULL || argv[arg + 1][0] == '-') {
                continue;
            }
            arg++;
            covG_filename[MAXFILELEN-1]='\0';
            strncpy(covG_filename, argv[arg], MAXFILELEN);
            if(covG_filename[MAXFILELEN-1]!='\0'){
                printf("ERROR: Input file names cannot exceed %d characters. Please double check the name of the ancestry-informative covariates file!\n", MAXFILELEN-1);
                exit(1);
            }
            cfile=1;
        } else if (strcmp(argv[arg], "--grm") == 0){
            if (argv[arg + 1] == NULL || argv[arg + 1][0] == '-') {
                continue;
            }
            arg++;
            grm_filename[MAXFILELEN-1]='\0';
            strncpy(grm_filename, argv[arg], MAXFILELEN);
            if(grm_filename[MAXFILELEN-1]!='\0'){
                printf("ERROR: Input file names cannot exceed %d characters. Please double check the name of the GRM matrix file!\n", MAXFILELEN-1);
                exit(1);
            }
            gfile=1;
        } else if (strcmp(argv[arg], "--eigen") == 0){
            if (argv[arg + 1] == NULL || argv[arg + 1][0] == '-') {
                continue;
            }
            arg++;
            eigen_filename[MAXFILELEN-1]='\0';
            strncpy(eigen_filename, argv[arg], MAXFILELEN);
            if(eigen_filename[MAXFILELEN-1]!='\0'){
                printf("ERROR: Input file names cannot exceed %d characters. Please double check the name of the eigen-decomposition file!\n", MAXFILELEN-1);
                exit(1);
            }
            efile=1;
        }   else if (strcmp(argv[arg], "-s") == 0){
            print_eigen = 1;
        } else if (strcmp(argv[arg], "--mc") == 0){
            if (argv[arg + 1] == NULL || argv[arg + 1][0] == '-') {
                continue;
            }
            arg++;
            errno = 0;
            niter_long = strtol(argv[arg], &end, 10);
            if(argv[arg] == end || errno != 0) {
                printf("ERROR: Cannot recognize the specified number of MC iterations!\n");
                exit(1);
            }
            if(niter_long > NITER_MAX || niter_long <= 0) {
                printf("ERROR: Number of MC iterations must be positive and less than %d.\n", (int) NITER_MAX);
                exit(1);
            } else niter = (int) niter_long;
        } else if (strcmp(argv[arg], "--out") == 0){
            if (argv[arg + 1] == NULL || argv[arg + 1][0] == '-') {
                continue;
            }
            arg++;
            ofile = 1;
            out_prefix[MAXFILELEN-1]='\0';
            strncpy(out_prefix, argv[arg], MAXFILELEN);
            if(out_prefix[MAXFILELEN-1]!='\0'){
                printf("ERROR: Input name cannot exceed %d characters. Modify option '--out'!\n", MAXFILELEN-1);
                exit(1);
            }
        } else if (strcmp(argv[arg], "-h") == 0 || strcmp(argv[arg], "--help") == 0){
            printf("Usage: JASPER --pheno --geno  --covG --grm --eigen -s --mc --out             \n"
                    "    pheno: name for trait file                                              \n"
                    "     geno: name for genotype file                                           \n"
                    "     covG: name for ancestry-informative covariates file                    \n"
                    "      grm: name for GRM file                                                \n"
                    "    eigen: name for GRM eigen-decomposition results binary file (optional)  \n"
                    "        s: save eigen-decomposition results in file 'grm.eigen' (optional)  \n"
                    "       mc: number of MC iterations if Pearson approximation fails           \n"
                    "      out: prefix for p-value output file                                   \n");
            exit(1);
        } else {
            printf("ERROR: Unrecognized option \"%s\" is ignored!\n", argv[arg]);
            continue;
        }
    }


    // Check all necessary arguments have been specified
    if( (pheno_filename != NULL) && (pheno_filename[0] == '\0')){
        printf("ERROR: Trait file has not been specified! (--pheno)\n");
        exit(1);
    } 
    if(!ofile){
        printf("ERROR: Prefix has not been specified! (--out)\n");
        exit(1);
    } 
    if( (geno_filename != NULL) && (geno_filename[0] == '\0')){
        printf("ERROR: Marker files prefix has not been specified!\n");
        exit(1);
    } 
    if( (covG_filename != NULL) && (covG_filename[0] == '\0')){
        printf("WARNING: No ancestry-informative covariates have been specified.\n");
    } 
    if( ((grm_filename != NULL) && (grm_filename[0] == '\0')) && ((eigen_filename != NULL) && (eigen_filename[0] == '\0')) ){
        printf("ERROR: Neither the GRM or the eigen-decomposition results file has been specified. At least one must be specified!\n");
        exit(1);
    } 
    if(gfile == 1 && efile == 1){
        printf("WARNING: Both the GRM or the eigen-decomposition results file has been specified. Eigen-decomposition file will be used.\n");
        gfile=0;
    }
    //////////////////////////////////////////////////////////////

    int i, ntotal;
    double pvalue;
    struct KER_MAT *trait, *genos, *covarG;
    struct COV_MAT *cov_mats;
    // Hash table for individual ID
    struct HASH hash;
    hash.ind2ind = kh_init(str);
    FILE *outFile;

    trait = (struct KER_MAT *) malloc(sizeof(struct KER_MAT) * 1);
    genos = (struct KER_MAT *) malloc(sizeof(struct KER_MAT) * 1);
    covarG = (struct KER_MAT *) malloc(sizeof(struct KER_MAT) * 1);
    cov_mats = (struct COV_MAT *) malloc(sizeof(struct COV_MAT) * 1);

    // Phenotype
    readpheno(pheno_filename, trait, &hash);
    ntotal = trait->nrow;

    // Genotype
    genos->nrow = ntotal;
    readgeno(geno_filename, genos, hash);

    // Ancestry-informative covariates
    covarG->nrow = ntotal;
    if(cfile == 1){
        readcovG(covG_filename, covarG, hash);
    } else {
        covarG->ncol = 1; // only an intercept
        covarG->mat = dmatrix(1, ntotal, 1, covarG->ncol);
        for(i = 1; i <= ntotal; i++) covarG->mat[i][1]=1;
    }
    // Get centering transformation J
    covarG->cenmat = dmatrix(1, ntotal, 1, ntotal);
    MapRowXd Xmat (&covarG->mat[1][1], ntotal, covarG->ncol);
    MapRowXd Jmat (&covarG->cenmat[1][1], ntotal, ntotal);
    Jmat = MatrixXd::Identity(ntotal, ntotal) - Xmat * (Xmat.transpose() * Xmat).colPivHouseholderQr().solve(Xmat.transpose());

    printf("\n"
            "## number of individuals = %d                                          \n"
            "## number of traits = %d                                               \n"
            "## number of genetic markers = %d                                      \n"
            "## number of ancestry-informative covariates (including intercept) = %d\n\n"
            , ntotal, trait->ncol, genos->ncol, covarG->ncol);
    // Info on MC iterations
    if(niter > 0) printf("*** User specified to perform %d Monte Carlo simulations when Pearson approximation fails.\n\n", niter);

    // GRM or its eigen-decomposition
    cov_mats->V = dmatrix(1, ntotal, 1, ntotal);
    cov_mats->D = dvector(1, ntotal);
    if(gfile == 1){
        cov_mats->phi = dmatrix(1, ntotal, 1, ntotal);
        fprintf(stderr, "Reading GRM entries given in file: %s.\nObtaining eigen-decomposition results... ", grm_filename);

        readkin(grm_filename, cov_mats->phi, ntotal, hash);
        get_eigendecomp(cov_mats, covarG);
        printf("Done.\n");

        // Print eigen-decomposition?
        if(print_eigen == 1){
            if(!ofile){
                strcpy(eigout_filename, "grm_eigen");
            } else{
                strcpy(eigout_filename, out_prefix);
                strcat(eigout_filename, "_eigen");
            }
            printf("User specified to print the GRM eigen-decomposition results in file: %s.\n", eigout_filename);

            writeBin_eigen(eigout_filename, cov_mats, ntotal);
        }
        free_dmatrix(cov_mats->phi, 1, ntotal, 1, ntotal);
    } else{
        printf("User specified to use the GRM eigen-decomposition results given in file: %s.\n", eigen_filename);

        readeigen(eigen_filename, cov_mats, ntotal, covarG->ncol); 
        if (!std::is_sorted(&cov_mats->D[1], &cov_mats->D[ntotal])){
            printf("ERROR: Eigenvalues given must be sorted in ascending order!\n"); 
            exit(1);
        }
    }

    if(!ofile){
        strcpy(pval_filename, "pvals_out.txt");
    } else{
        strcpy(pval_filename, out_prefix);
        strcat(pval_filename, "_out.txt");
    }
    fprintf(stderr, "Printing p-value to file: %s.\n", pval_filename);
    pvalue = get_pval(trait, genos, covarG, cov_mats, niter);

    outFile = fopen(pval_filename, "w");
    fprintf(outFile, "pvalue %.17g\n", pvalue);
    fclose(outFile);

    // Free used arrays
    free_dmatrix(trait->mat, 1, ntotal, 1, trait->ncol);
    free_dmatrix(genos->mat, 1, ntotal, 1, genos->ncol);
    free_dmatrix(covarG->mat, 1, ntotal, 1, covarG->ncol);
    free_dmatrix(covarG->cenmat, 1, ntotal, 1, ntotal);
    free_dmatrix(cov_mats->V, 1, ntotal, 1, ntotal);
    free_dvector(cov_mats->D, 1, ntotal);

    free(trait);
    free(genos);
    free(covarG);
    free(cov_mats);

    for (int k = kh_begin(hash.ind2ind); k != kh_end(hash.ind2ind); k++){
        if (kh_exist(hash.ind2ind,k)) {
            free((char*) kh_key(hash.ind2ind,k)); /* cast away constness */
        }
    }
    kh_destroy(str,hash.ind2ind);

    time2 = clock();
    time_JASPER = (double) (time2 - time1) / CLOCKS_PER_SEC;

    printf("*** INFO: ");
    if(time_JASPER < 60){
        printf("JASPER took %.2lf seconds.\n", time_JASPER);
    } else{
        int mins_jasp = time_JASPER / 60;
        printf("JASPER took %d minutes and %.2lf seconds.\n", mins_jasp, time_JASPER - mins_jasp * 60);
    }

    return 1;
} /* end main */

void get_eigendecomp(struct COV_MAT *cov_mats, struct KER_MAT* cmat){

    int ntotal = cmat->nrow;
    double **cenmat = cmat->cenmat;
    double **grm = cov_mats->phi;
    double **evecs = cov_mats->V;
    double *evals = cov_mats->D;

    MapRowXd Jmat (&cenmat[1][1], ntotal, ntotal);
    MapRowXd GRMmap (&grm[1][1], ntotal, ntotal);
    MapRowXd Vmap (&evecs[1][1], ntotal, ntotal);
    MapArXd Dmap (&evals[1], ntotal, 1);

    MatrixXd JGJ = Jmat * GRMmap * Jmat;

    SelfAdjointEigenSolver<MatrixXd> eigGRM (JGJ);
    Vmap = eigGRM.eigenvectors();
    Dmap = eigGRM.eigenvalues().array(); // sorted in increasing order

}

void writeBin_eigen(char *fileout, struct COV_MAT *cov_mats, int ntotal){

    int i, j;
    FILE *efile;

    double **evecs = cov_mats->V;
    double *evals = cov_mats->D;

    efile = fopen(fileout, "wb");
    fwrite(&evals[1], sizeof(double), ntotal, efile);
    for(j = 1; j <= ntotal; j++)
        for(i = 1; i <= ntotal; i++)
            fwrite(&evecs[i][j], sizeof(double), 1, efile);
    fclose(efile);

}

double get_pval(struct KER_MAT *Y, struct KER_MAT *G, struct KER_MAT *covG, struct COV_MAT *cov_mats, int niter) {

    int ntotal = Y->nrow;
    int npheno = Y->ncol;
    int nmarker = G->ncol;
    int ncovG = covG->ncol;

    // Matrices needed
    MapRowXd Ymap (&Y->mat[1][1], ntotal, npheno); 
    MapRowXd Gmap (&G->mat[1][1], ntotal, nmarker); 
    MapRowXd Jmap (&covG->cenmat[1][1], ntotal, ntotal); 
    MapRowXd Vmap (&cov_mats->V[1][1], ntotal, ntotal); 
    MapArXd Dmap (&cov_mats->D[1], ntotal, 1); 


    // Last (n-R) eigenvalues are non-zero (where R = # ancestry-informative covariates)
    ArrayXd Dpos_sqrt = Dmap.bottomRows(ntotal - ncovG).sqrt(); 
    MatrixXd Vpos = Vmap.rightCols(ntotal - ncovG);

    // Transformed phenotype and genotype matrices
    MatrixXd Ynew = Dpos_sqrt.matrix().asDiagonal() * Vpos.transpose() * Ymap;
    MatrixXd Gnew = (1/Dpos_sqrt).matrix().asDiagonal() * Vpos.transpose() * Jmap * Gmap;


    // Get p-value
    return pearson_pval(Ynew, Gnew, niter);

}

// Using Pearson type III approx. with parameters :shape,scale,location
double pearson_pval(const Ref<const MatrixXd>& Ymat, const Ref<const MatrixXd>& Gmat, int niter){

    double Q, pv, Z, location, shape, scale;
    int  sign_Z, sign_scale;
    struct PARAMS results;

    MatrixXd Yker = Ymat * Ymat.transpose();
    MatrixXd Gker = Gmat * Gmat.transpose();

    Q = Yker.cwiseProduct(Gker).sum();
    get_moments(Yker, Gker, &results);

    scale = results.sc;
    shape = results.sh;
    location = results.loc; 

    Z = Q - location;
    sign_Z = sign(Z);
    sign_scale = sign(scale); 

    boost::math::gamma_distribution<double> my_gamma(shape, 1);

    if(sign_scale < 0 && sign_Z <= 0){
        pv = cdf(my_gamma, Z/scale);
    } else if(sign_scale > 0 && sign_Z >= 0) {
        pv = cdf(complement(my_gamma, Z/scale));
    } else {
        printf("WARNING: Pearson approximation failed. ");
        if(niter > 0 || sign_scale == 0){
            pv = mc_iter(Yker, Gmat, Q, niter);
        } else {
            if(sign_scale < 0){
                printf("Setting p-value to 0.\n");
                pv = 0;
            } else {
                printf("Setting p-value to 1.\n");
                pv = 1;
            }
        }
    }

    return pv;
}

// Monte Carlo simulations to permute the rows of Aker
double mc_iter(const Ref<const MatrixXd>& Yker, const Ref<const MatrixXd>& Gmat, double Qobs, int n_iter){

    int i, pval;
    double Qperm;
    MatrixXd Gperm, Gker;

    printf("Using %d Monte Carlo simulations to estimate the p-value.\n", n_iter);

    PermutationMatrix<Dynamic,Dynamic> perm(Gmat.rows());
    perm.setIdentity();

    for(pval = 1, i = 1; i <= n_iter; i++){

        // Permute rows
        std::random_shuffle(perm.indices().data(), perm.indices().data()+perm.indices().size());
        Gperm = perm * Gmat;
        // Get new Q and compare to Qobs
        Gker = Gperm * Gperm.transpose();
        Qperm = Gker.cwiseProduct(Yker).sum();
        if(Qperm >= Qobs) pval++;
    }

    return (1.0 * pval) / (n_iter + 1);
}

void get_moments(const Ref<const MatrixXd>& Aker, const Ref<const MatrixXd>& Wker, struct PARAMS *params){

    struct QUANTS quantA, quantW;

    int nrow = Aker.rows();
    if(nrow <= 5){
        printf("Number of rows in transformed phenotype & genotype matrices is too small (=%d)! It must be at least 6.\n", nrow);
        exit(1);
    }

    get_quantities(Aker, &quantA);
    get_quantities(Wker, &quantW);

    long double Aii, Aii2, Aii3, Aij, Aij2, AiiAij, AiiAjj, AiiAjjAij, AiiAij2, Aii2Aij, Aii2Ajj, Aij3, AiiAjk, AijAik, AiiAjjAkk, AiiAjjAik, AiiAijAik, AiiAijAjk, AiiAjk2, AijAikAjk, Aii2Ajk, Aij2Aik, AijAkl, AiiAjjAkl, AiiAijAkl, AiiAjkAjl, AijAikAil, AijAikAjl, Aij2Akl, AijAikAlm, AiiAjkAlm, AijAklAmn;
    long double Wii, Wii2, Wii3, Wij, Wij2, WiiWij, WiiWjj, WiiWjjWij, WiiWij2, Wii2Wij, Wii2Wjj, Wij3, WiiWjk, WijWik, WiiWjjWkk, WiiWjjWik, WiiWijWik, WiiWijWjk, WiiWjk2, WijWikWjk, Wii2Wjk, Wij2Wik, WijWkl, WiiWjjWkl, WiiWijWkl, WiiWjkWjl, WijWikWil, WijWikWjl, Wij2Wkl, WijWikWlm, WiiWjkWlm, WijWklWmn;

    // Single index
    Wii = quantW.T1;
    Wii2 = quantW.T2;
    Wii3 = quantW.T3;
    /////////
    Aii = quantA.T1;
    Aii2 = quantA.T2;
    Aii3 = quantA.T3;


    // Two indices
    Wij = quantW.S1 - quantW.T1;
    Wij2 = quantW.S2 - quantW.T2;
    Wij3 = quantW.S3 - quantW.T3;
    WiiWij = quantW.R1 - quantW.T2;
    WiiWij2 = quantW.R2 - quantW.T3;
    Wii2Wij = quantW.R3 - quantW.T3;
    WiiWjj = quantW.T1 * quantW.T1 - quantW.T2;
    Wii2Wjj = quantW.T2 * quantW.T1 - quantW.T3;
    WiiWjjWij = quantW.R4 - quantW.T3;
    /////////
    Aij = quantA.S1 - quantA.T1;
    Aij2 = quantA.S2 - quantA.T2;
    Aij3 = quantA.S3 - quantA.T3;
    AiiAij = quantA.R1 - quantA.T2;
    AiiAij2 = quantA.R2 - quantA.T3;
    Aii2Aij = quantA.R3 - quantA.T3;
    AiiAjj = quantA.T1 * quantA.T1 - quantA.T2;
    Aii2Ajj = quantA.T2 * quantA.T1 - quantA.T3;
    AiiAjjAij = quantA.R4 - quantA.T3;


    // Three indices 
    WiiWjk = quantW.T1*(quantW.S1 - quantW.T1) - 2*(quantW.R1 - quantW.T2);
    WiiWjk2 = quantW.T1*(quantW.S2 - quantW.T2) - 2* (quantW.R2 - quantW.T3);
    Wii2Wjk = quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3);
    WijWik = quantW.C1-quantW.R1-(quantW.R1 - quantW.T2)-(quantW.S2 - quantW.T2);
    Wij2Wik  = quantW.C2-quantW.R3-(quantW.S3 - quantW.T3)-(quantW.R2 - quantW.T3);
    WiiWijWjk = quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3);
    WiiWijWik = quantW.C3-quantW.R3-(quantW.R3 - quantW.T3)-(quantW.R2 - quantW.T3);
    WiiWjjWik = quantW.T1*(quantW.R1 - quantW.T2)-(quantW.R4 - quantW.T3)-(quantW.R3 - quantW.T3);
    WiiWjjWkk = quantW.T1*(quantW.T1 * quantW.T1 - quantW.T2)-2*(quantW.T2 * quantW.T1 - quantW.T3);
    WijWikWjk = quantW.C5-quantW.R2-2*(quantW.R2 - quantW.T3);
    /////////
    AiiAjk = quantA.T1*(quantA.S1 - quantA.T1) - 2*(quantA.R1 - quantA.T2);
    AiiAjk2 = quantA.T1*(quantA.S2 - quantA.T2) - 2* (quantA.R2 - quantA.T3);
    Aii2Ajk = quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3);
    AijAik = quantA.C1-quantA.R1-(quantA.R1 - quantA.T2)-(quantA.S2 - quantA.T2);
    Aij2Aik  = quantA.C2-quantA.R3-(quantA.S3 - quantA.T3)-(quantA.R2 - quantA.T3);
    AiiAijAjk = quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3);
    AiiAijAik = quantA.C3-quantA.R3-(quantA.R3 - quantA.T3)-(quantA.R2 - quantA.T3);
    AiiAjjAik = quantA.T1*(quantA.R1 - quantA.T2)-(quantA.R4 - quantA.T3)-(quantA.R3 - quantA.T3);
    AiiAjjAkk = quantA.T1*(quantA.T1 * quantA.T1 - quantA.T2)-2*(quantA.T2 * quantA.T1 - quantA.T3);
    AijAikAjk = quantA.C5-quantA.R2-2*(quantA.R2 - quantA.T3);


    // Four indices 
    WijWkl = quantW.S1 * (quantW.S1 - quantW.T1) + 2*(quantW.R1-quantW.C1) - 2*(quantW.C1-quantW.R1-(quantW.R1 - quantW.T2)-(quantW.S2 - quantW.T2)) - (quantW.T1*(quantW.S1 - quantW.T1) - 2*(quantW.R1 - quantW.T2));
    Wij2Wkl = quantW.S2*(quantW.S1 - quantW.T1)+2*(quantW.R2-quantW.C2)-2*(quantW.C2-quantW.R3-(quantW.S3 - quantW.T3)-(quantW.R2 - quantW.T3))-(quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3));
    WiiWjjWkl = quantW.T1*(quantW.T1*(quantW.S1 - quantW.T1) - 2*(quantW.R1 - quantW.T2)) - (quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3)) - 2*(quantW.T1*(quantW.R1 - quantW.T2)-(quantW.R4 - quantW.T3)-(quantW.R3 - quantW.T3));
    WiiWijWkl = quantW.R1*(quantW.S1 - quantW.T1)+2*(quantW.R3-quantW.C3)-2*(quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3))-(quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3));
    WiiWjkWjl = quantW.T1 * (quantW.C1-quantW.R1-(quantW.R1 - quantW.T2)-(quantW.S2 - quantW.T2)) - (quantW.C3-quantW.R3-(quantW.R3 - quantW.T3)-(quantW.R2 - quantW.T3)) - 2*(quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3));
    WijWikWil = quantW.D1-3*quantW.C2+2*quantW.S3-3*(quantW.C3-quantW.R3-(quantW.R3 - quantW.T3)-(quantW.R2 - quantW.T3));
    WijWikWjl = quantW.D2+2*quantW.R2-(quantW.C2+quantW.C4+quantW.C5)-(quantW.C2-quantW.R3-(quantW.S3 - quantW.T3)-(quantW.R2 - quantW.T3))-(quantW.C3-quantW.R3-(quantW.R3 - quantW.T3)-(quantW.R2 - quantW.T3))-(quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3));
    /////////
    AijAkl = quantA.S1 * (quantA.S1 - quantA.T1) + 2*(quantA.R1-quantA.C1) - 2*(quantA.C1-quantA.R1-(quantA.R1 - quantA.T2)-(quantA.S2 - quantA.T2)) - (quantA.T1*(quantA.S1 - quantA.T1) - 2*(quantA.R1 - quantA.T2));
    Aij2Akl = quantA.S2*(quantA.S1 - quantA.T1)+2*(quantA.R2-quantA.C2)-2*(quantA.C2-quantA.R3-(quantA.S3 - quantA.T3)-(quantA.R2 - quantA.T3))-(quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3));
    AiiAjjAkl = quantA.T1*(quantA.T1*(quantA.S1 - quantA.T1) - 2*(quantA.R1 - quantA.T2)) - (quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3)) - 2*(quantA.T1*(quantA.R1 - quantA.T2)-(quantA.R4 - quantA.T3)-(quantA.R3 - quantA.T3));
    AiiAijAkl = quantA.R1*(quantA.S1 - quantA.T1)+2*(quantA.R3-quantA.C3)-2*(quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3))-(quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3));
    AiiAjkAjl = quantA.T1 * (quantA.C1-quantA.R1-(quantA.R1 - quantA.T2)-(quantA.S2 - quantA.T2)) - (quantA.C3-quantA.R3-(quantA.R3 - quantA.T3)-(quantA.R2 - quantA.T3)) - 2*(quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3));
    AijAikAil = quantA.D1-3*quantA.C2+2*quantA.S3-3*(quantA.C3-quantA.R3-(quantA.R3 - quantA.T3)-(quantA.R2 - quantA.T3));
    AijAikAjl = quantA.D2+2*quantA.R2-(quantA.C2+quantA.C4+quantA.C5)-(quantA.C2-quantA.R3-(quantA.S3 - quantA.T3)-(quantA.R2 - quantA.T3))-(quantA.C3-quantA.R3-(quantA.R3 - quantA.T3)-(quantA.R2 - quantA.T3))-(quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3));


    // Five indices 
    WiiWjkWlm = quantW.T1 * (quantW.S1 * (quantW.S1 - quantW.T1) + 2*(quantW.R1-quantW.C1) - 2*(quantW.C1-quantW.R1-(quantW.R1 - quantW.T2)-(quantW.S2 - quantW.T2)) - (quantW.T1*(quantW.S1 - quantW.T1) - 2*(quantW.R1 - quantW.T2))) - 4*(quantW.R1*(quantW.S1 - quantW.T1)+2*(quantW.R3-quantW.C3)-2*(quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3))-(quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3)));
    WijWikWlm = (quantW.S1 - quantW.T1) * (quantW.C1 - quantW.S2) + 4*(quantW.C2+quantW.C4-quantW.D2)+2*quantW.C5-6*quantW.R2-2*((quantW.R1*(quantW.S1 - quantW.T1)+2*(quantW.R3-quantW.C3)-2*(quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3))-(quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3)))+(quantW.D1-3*quantW.C2+2*quantW.S3-3*(quantW.C3-quantW.R3-(quantW.R3 - quantW.T3)-(quantW.R2 - quantW.T3))));
    /////////
    AiiAjkAlm = quantA.T1 * (quantA.S1 * (quantA.S1 - quantA.T1) + 2*(quantA.R1-quantA.C1) - 2*(quantA.C1-quantA.R1-(quantA.R1 - quantA.T2)-(quantA.S2 - quantA.T2)) - (quantA.T1*(quantA.S1 - quantA.T1) - 2*(quantA.R1 - quantA.T2))) - 4*(quantA.R1*(quantA.S1 - quantA.T1)+2*(quantA.R3-quantA.C3)-2*(quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3))-(quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3)));
    AijAikAlm = (quantA.S1 - quantA.T1) * (quantA.C1 - quantA.S2) + 4*(quantA.C2+quantA.C4-quantA.D2)+2*quantA.C5-6*quantA.R2-2*((quantA.R1*(quantA.S1 - quantA.T1)+2*(quantA.R3-quantA.C3)-2*(quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3))-(quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3)))+(quantA.D1-3*quantA.C2+2*quantA.S3-3*(quantA.C3-quantA.R3-(quantA.R3 - quantA.T3)-(quantA.R2 - quantA.T3))));


    // Six indices 
    WijWklWmn = quantW.S1*(quantW.S1 * (quantW.S1 - quantW.T1) + 2*(quantW.R1-quantW.C1) - 2*(quantW.C1-quantW.R1-(quantW.R1 - quantW.T2)-(quantW.S2 - quantW.T2)) - (quantW.T1*(quantW.S1 - quantW.T1) - 2*(quantW.R1 - quantW.T2)))+4*(quantW.R1-quantW.C1)*(quantW.S1 - quantW.T1)+8*( quantW.D1+quantW.D2-quantW.C2-quantW.C4+3*(quantW.R3-quantW.C3) ) -4*((quantW.S1 - quantW.T1) * (quantW.C1 - quantW.S2) + 4*(quantW.C2+quantW.C4-quantW.D2)+2*quantW.C5-6*quantW.R2-2*((quantW.R1*(quantW.S1 - quantW.T1)+2*(quantW.R3-quantW.C3)-2*(quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3))-(quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3)))+(quantW.D1-3*quantW.C2+2*quantW.S3-3*(quantW.C3-quantW.R3-(quantW.R3 - quantW.T3)-(quantW.R2 - quantW.T3)))))-(quantW.T1 * (quantW.S1 * (quantW.S1 - quantW.T1) + 2*(quantW.R1-quantW.C1) - 2*(quantW.C1-quantW.R1-(quantW.R1 - quantW.T2)-(quantW.S2 - quantW.T2)) - (quantW.T1*(quantW.S1 - quantW.T1) - 2*(quantW.R1 - quantW.T2))) - 4*(quantW.R1*(quantW.S1 - quantW.T1)+2*(quantW.R3-quantW.C3)-2*(quantW.C4-quantW.R3-(quantW.R2 - quantW.T3)-(quantW.R4 - quantW.T3))-(quantW.T2*(quantW.S1 - quantW.T1)-2*(quantW.R3 - quantW.T3))));
    /////////
    AijAklAmn = quantA.S1*(quantA.S1 * (quantA.S1 - quantA.T1) + 2*(quantA.R1-quantA.C1) - 2*(quantA.C1-quantA.R1-(quantA.R1 - quantA.T2)-(quantA.S2 - quantA.T2)) - (quantA.T1*(quantA.S1 - quantA.T1) - 2*(quantA.R1 - quantA.T2)))+4*(quantA.R1-quantA.C1)*(quantA.S1 - quantA.T1)+8*( quantA.D1+quantA.D2-quantA.C2-quantA.C4+3*(quantA.R3-quantA.C3) ) -4*((quantA.S1 - quantA.T1) * (quantA.C1 - quantA.S2) + 4*(quantA.C2+quantA.C4-quantA.D2)+2*quantA.C5-6*quantA.R2-2*((quantA.R1*(quantA.S1 - quantA.T1)+2*(quantA.R3-quantA.C3)-2*(quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3))-(quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3)))+(quantA.D1-3*quantA.C2+2*quantA.S3-3*(quantA.C3-quantA.R3-(quantA.R3 - quantA.T3)-(quantA.R2 - quantA.T3)))))-(quantA.T1 * (quantA.S1 * (quantA.S1 - quantA.T1) + 2*(quantA.R1-quantA.C1) - 2*(quantA.C1-quantA.R1-(quantA.R1 - quantA.T2)-(quantA.S2 - quantA.T2)) - (quantA.T1*(quantA.S1 - quantA.T1) - 2*(quantA.R1 - quantA.T2))) - 4*(quantA.R1*(quantA.S1 - quantA.T1)+2*(quantA.R3-quantA.C3)-2*(quantA.C4-quantA.R3-(quantA.R2 - quantA.T3)-(quantA.R4 - quantA.T3))-(quantA.T2*(quantA.S1 - quantA.T1)-2*(quantA.R3 - quantA.T3))));


    long double n = Aker.rows();
    long double d1, d2, d3, d4, d5, d6, t1, t2, t3, t4, t5, t6;
    long double EQ, EQ2, EQ3;

    // For EQ
    d1 = n;
    d2 = n * (n-1.0);

    t1 = Aii * Wii / d1;
    t2 = Aij * Wij / d2;

    EQ = t1 + t2;
    /////////

    // For EQ2
    d3 = n*(n-1.0)*(n-2.0);
    d4 = n*(n-1.0)*(n-2.0)*(n-3.0);

    t1 = Aii2 * Wii2 / d1;
    t2 = (4 * AiiAij * WiiWij + 2 * Aij2 * Wij2 + AiiAjj * WiiWjj) /d2 ;
    t3 = (4 * AijAik * WijWik + 2 * AiiAjk * WiiWjk) / d3 ;
    t4 = AijAkl * WijWkl / d4 ;

    EQ2 = t1 + t2 + t3 + t4;
    /////////

    // For EQ3
    d5 = n*(n-1.0)*(n-2.0)*(n-3.0)*(n-4.0);
    d6 = n*(n-1.0)*(n-2.0)*(n-3.0)*(n-4.0)*(n-5.0);

    t1 = Aii3 * Wii3 / d1;
    t2 = (6* AiiAjjAij * WiiWjjWij + 12* AiiAij2 * WiiWij2 + 6* Aii2Aij * Wii2Wij + 3* Aii2Ajj * Wii2Wjj + 4* Aij3 * Wij3) / d2;
    t3 = (AiiAjjAkk * WiiWjjWkk + 12* AiiAjjAik * WiiWjjWik + 12* AiiAijAik * WiiWijWik + 24* AiiAijAjk * WiiWijWjk + 6* AiiAjk2 * WiiWjk2 + 8* AijAikAjk * WijWikWjk + 3 * Aii2Ajk * Wii2Wjk + 24* Aij2Aik * Wij2Wik) / d3;
    t4 = (3* AiiAjjAkl * WiiWjjWkl + 12* AiiAijAkl* WiiWijWkl + 12* AiiAjkAjl * WiiWjkWjl + 8* AijAikAil * WijWikWil + 24* AijAikAjl * WijWikWjl + 6* Aij2Akl * Wij2Wkl) / d4;
    t5 = (12* AijAikAlm * WijWikWlm + 3* AiiAjkAlm * WiiWjkWlm) / d5;
    t6 = AijAklAmn * WijWklWmn / d6;

    EQ3 = t1 + t2 + t3 + t4 + t5 + t6;

    /////////
    double varQ, sdQ, skewQ;

    varQ = EQ2 - EQ * EQ;
    sdQ = sqrt(varQ);
    skewQ = (EQ3 - 3.0 * EQ * varQ - EQ * EQ * EQ) / (sdQ * sdQ * sdQ);

    params->sc = sdQ * skewQ / 2.0;
    params->sh = 4.0 / (skewQ * skewQ);
    params->loc = EQ - 2.0 * sdQ / skewQ;
    params->EQ = EQ;
    params->EQ2 = EQ2;
    params->EQ3 = EQ3;

}

void get_quantities(const Ref<const MatrixXd>& A, struct QUANTS* qA){

    VectorXd dd = A.diagonal();
    VectorXd dd_sq = dd.array().square();
    VectorXd A_rsum = A.rowwise().sum();
    VectorXd A_rsum_sq = A_rsum.array().square();
    MatrixXd A_sq = A.array().square().matrix();
    VectorXd A_sq_rsum = A_sq.rowwise().sum();
    VectorXd A_A_rsum = A * A_rsum;


    qA->T1 = dd.sum();                     // Aii
    qA->T2 = dd_sq.sum();                  // Aii2
    qA->T3 = dd.array().cube().sum();      // Aii3
    qA->S1 = A.sum();                      // Aij
    qA->S2 = A_sq.sum();                   // Aij2
    qA->S3 = A.array().cube().sum();       // Aij3
    qA->R1 = dd.dot(A_rsum);               // AiiAij
    qA->R2 = dd.dot(A_sq_rsum);            // AiiAij2
    qA->R3 = dd_sq.dot(A_rsum);            // Aii2Aij
    qA->R4 = (A * dd).dot(dd);              // AiiAjjAij
    qA->C1 = A_rsum_sq.sum();               // AijAik
    qA->C2 = A_sq_rsum.dot(A_rsum);        // Aij2Aik
    qA->C3 = dd.dot(A_rsum_sq);            // AiiAijAik
    qA->C4 = dd.dot(A_A_rsum);             // AiiAijAjk
    qA->C5 = A.cwiseProduct(A*A).sum();    // AijAikAjk
    qA->D1 = A_rsum.array().cube().sum();  // AijAikAil
    qA->D2 = A_rsum.dot(A_A_rsum);         // AijAikAjl

}


