#include "read.h"

// For phenotype
void readpheno(char *filename, struct KER_MAT* tmat, struct HASH *hash){

    int i, j, col = 0, sex, ret;
    double pheno;
    char car, fam[MAXIDLEN], ind_id[MAXIDLEN], fa[MAXIDLEN], mo[MAXIDLEN], *end = NULL;
    FILE *oFile;

    oFile = fopen(filename, "r");
    if(oFile == NULL){
        printf("Can't open file: %s!\n", filename);
        exit(1);
    }

    /* read through the file to count the number of columns to then find the number of traits */
    car = 0;
    while (car != EOF && car != '\n') {
        car = getc(oFile);
        if (car == ' ' || car == '\t') {
            col++;
        }
    }
    col++;
    tmat->ncol = col - 5;
    fclose(oFile);

    // Count number of individuals in file
    tmat->nrow = 0;
    oFile = fopen(filename, "r");
    while (!feof(oFile)) {
        if( fscanf(oFile, "%d %s %s %s %d ", &fam, &ind_id, &fa, &mo, &sex) != 5) {
            printf("ERROR: Check format of phenotype file!\n");
            exit(1);
        }
        // Check all individuals have the same number of traits
        for(i = 1; i < tmat->ncol; i++)
            if( fscanf(oFile, "%lf ", &pheno) != 1) {
                printf("ERROR: Check format of phenotype file!\n");
                exit(1);
            }
        if( fscanf(oFile, "%lf\n", &pheno) != 1) {
            printf("ERROR: Check format of phenotype file!\n");
            exit(1);
        }
        tmat->nrow++;
    }
    fclose(oFile);

    tmat->mat = dmatrix(1, tmat->nrow, 1, tmat->ncol);
    // Initialize hash table
    if (hash->ind2ind != NULL) kh_destroy(str, hash->ind2ind);
    hash->ind2ind = kh_init(str);
    khiter_t k_i;

    // Save phenotype info
    oFile = fopen(filename, "r");
    i = 0;
    while (!feof(oFile)) {
        fscanf(oFile, "%d %s %s %s %d ", &fam, ind_id, &fa, &mo, &sex);

        // Fill hash table
        k_i = kh_put(str, hash->ind2ind, strdup(ind_id), &ret);
        kh_value(hash->ind2ind, k_i) = ++i;

        for(j = 1; j < tmat->ncol; j++){
            fscanf(oFile, "%lf ", &pheno);
            tmat->mat[i][j] = pheno;
        }
        fscanf(oFile, "%lf\n", &pheno);
        tmat->mat[i][j] = pheno;
    }
    fclose(oFile);


}

// For ancestry-informative covariates 
void readcovG(char *filename, struct KER_MAT* cmat, struct HASH hash){

    int i, j, col = 0, ind;
    double cov;
    char car, fam[MAXIDLEN], ind_id[MAXIDLEN];
    khiter_t k_i;
    FILE *oFile;

    oFile = fopen(filename, "r");
    if(oFile == NULL){
        printf("Can't open file: %s!\n", filename);
        exit(1);
    }

    /* read through the file to count the number of columns to then find the number of ancestry-informative covariates */
    car = 0;
    while (car != EOF && car != '\n') {
        car = getc(oFile);
        if (car == ' ' || car == '\t') {
            col++;
        }
    }
    col++;
    cmat->ncol = col - 2 + 1; // Add intercept column
    if(cmat->ncol < 1) {
        printf("ERROR: Check format of ancestry-informative covariates file: %s!\n", filename);
        printf("If no ancestry-informative covariates need to be provided, omit option '--covG'.\n");
        exit(1);
    }
    fclose(oFile);

    cmat->mat = dmatrix(1, cmat->nrow, 1, cmat->ncol);
    // Save covariate info
    oFile = fopen(filename, "r");
    i = 0;

    while (!feof(oFile)) {
        if(fscanf(oFile, "%s %s ", &fam, ind_id) != 2){
            printf("ERROR: Check format of file %s!\n", filename);
            exit(1);
        }
        k_i = kh_get(str, hash.ind2ind, ind_id);
        ind = kh_value(hash.ind2ind, k_i);

        if(++i != ind){
            printf("Order of individuals in file %s must be the same as in the phenotype file!\n", filename);
            exit(1);
        }

        cmat->mat[i][1] = 1;
        for(j = 2; j < cmat->ncol; j++){
            fscanf(oFile, "%lf ", &cov);
            cmat->mat[i][j] = cov;
        }
        fscanf(oFile, "%lf\n", &cov);
        cmat->mat[i][j] = cov;
    }
    fclose(oFile);

}



// For genotype file
void readgeno(char *filename, struct KER_MAT* gmat, struct HASH hash){

    int i, j, ind, line, col = 0;
    double geno;
    char car, snp_id[MAXIDLEN], ind_id[MAXIDLEN];
    khiter_t k_i;
    FILE *oFile;

    oFile = fopen(filename, "r");
    if(oFile == NULL){
        printf("Can't open file: %s!\n", filename);
        exit(1);
    }

    /* Read through the file to count the number of rows (=the number of markers) */
    line = 0;
    while (!feof(oFile)) {
        ++line;

        if( fscanf(oFile, "%s ", snp_id) != 1) {
            printf("ERROR: Check format of genotype file!\n");
            exit(1);
        }
        for(i = 1; i <= gmat->nrow; i++)
            // First line is individuals ID
            if (line == 1){
                if( fscanf(oFile, i == gmat->nrow? "%s\n" : "%s ", ind_id) != 1) {
                    printf("ERROR: Check format of genotype file!\n");
                    exit(1);
                }
                // Check order of individuals
                k_i = kh_get(str, hash.ind2ind, ind_id);
                ind = kh_value(hash.ind2ind, k_i);

                if( i != ind){
                    printf("Order of individuals in file %s must be the same as in the phenotype file!\n", filename);
                    exit(1);
                }
            } else {
                // Other lines are genotypes
                if( fscanf(oFile, i == gmat->nrow? "%lf\n" : "%lf ", &geno) != 1) {
                    printf("ERROR: Check format of genotype file!\n");
                    exit(1);
                }
                if(geno < 0 || geno > 2){
                    printf("ERROR: Genotypes must be in [0,2]!\n");
                    exit(1);
                }
            }
    }
    fclose(oFile);
    gmat->ncol = line - 1;

    gmat->mat = dmatrix(1, gmat->nrow, 1, gmat->ncol);
    // Save genotype info
    oFile = fopen(filename, "r");
    // Skip first line
    car= fgetc(oFile);
    while (car != '\n')
        car= fgetc(oFile);

    line = 0;
    while (!feof(oFile)) {
        ++line;

        fscanf(oFile, "%s ", snp_id);

        for(i = 1; i <= gmat->nrow; i++){
            fscanf(oFile, i == gmat->nrow? "%lf\n" : "%lf ", &geno); 
            gmat->mat[i][line] = geno;
        }
    }
    fclose(oFile);

}


// Read in the GRM
void readkin(char *filename, double **grm, int ntot, struct HASH hash){

    int ind1, ind2;
    double kcoef;
    char ind1_id[MAXIDLEN], ind2_id[MAXIDLEN];
    khiter_t k_i;

    FILE *kinfile;
    kinfile = fopen(filename, "r");
    if (kinfile == NULL) {
        printf("ERROR: Can't open GRM file: %s!\n", filename);
        exit(1);
    }

    // Initialize GRM with missing entries
    for (ind1 = 1; ind1 <= ntot ; ++ind1) {
        for (ind2 = 1; ind2 < ind1 ; ++ind2)
            grm[ind1][ind2] = grm[ind2][ind1] = -9;
        grm[ind1][ind1] = -9;
    }

    // Store GRM
    while (!feof(kinfile)) {
        if (fscanf(kinfile, "%s %s %lf\n", ind1_id, ind2_id, &kcoef) != 3) exit(1);

        // Get indices from hash table
        k_i = kh_get(str, hash.ind2ind, ind1_id);
        ind1 = kh_value(hash.ind2ind, k_i);

        k_i = kh_get(str, hash.ind2ind, ind2_id);
        ind2 = kh_value(hash.ind2ind, k_i);

        // Check if GRM entry is already filled (if it is, discard the new value)
        if(ind1 == ind2) {
            if(grm[ind1][ind1]!=-9){
                printf("WARNING: In GRM file, the diagonal element for individual %s is included more than once! The new value is discarded.\n", ind1_id);
            } else grm[ind1][ind1] = kcoef;
        } else {
            if(grm[ind1][ind2]!=-9){
                printf("WARNING: In GRM file, the individual pair (%s and %s) is included more than once! The new value is discarded.\n", ind1_id, ind2_id);
            } else grm[ind1][ind2] = grm[ind2][ind1] = kcoef;
        }
    }
    fclose(kinfile);

    // Fill in remaning entries with 0/1 if missing
    for (ind1 = 1; ind1 <= ntot ; ++ind1) {
        for (ind2 = 1; ind2 < ind1 ; ++ind2)
            if(grm[ind1][ind2] == -9){
                grm[ind1][ind2] = grm[ind2][ind1] = 0;
            }
        if(grm[ind1][ind1] == -9) grm[ind1][ind1] = 1;
    }
}

// Read eigen-decomposition binary file (eigenvalues then corresponding eigenvectors)
void readeigen(char *filename, struct COV_MAT *cov_mats, int ntotal, int nzero){

    int i, j;
    FILE *oFile;

    double **evecs = cov_mats->V;
    double *evals = cov_mats->D;

    oFile = fopen(filename, "rb");
    if(oFile == NULL){
        printf("Can't open file: %s!\n", filename);
        exit(1);
    }

    // Read eigenvalues
    if(fread(&evals[1], sizeof(double), ntotal, oFile) != ntotal){
        printf("Error while reading file: %s!\n", filename);
        exit(1);
    }
    for(j = 0, i = 1; i<= ntotal; i++){

        if(evals[i] < -NUMTOL){
            printf("ERROR: Negative eigenvalues are not allowed in file %s!\n", filename);
            exit(1);
        }
        if(evals[i] < NUMTOL) j++;
    }
    if( j != nzero){
        printf("ERROR: Number of zero eigenvalues is not equal to number of ancestry-informative covariates (including intercept). Eigen-decomposition results must be obtained from centered GRM (using ancestry-informative covariates).\n");
        exit(1);
    }

    for(j = 1; j <= ntotal; j++)
        for(i = 1; i <= ntotal; i++)
            if(fread(&evecs[i][j], sizeof(double), 1, oFile) != 1){
                printf("Error while reading file: %s!\n", filename);
                exit(1);
            }

    fclose(oFile);

}

